@@ -1,18 +1,19 @@
 """ KNN retrieval using an Annoy index. """
 
+from concurrent.futures import ThreadPoolExecutor
 import os
 import shutil
 import time
 from abc import abstractmethod
 from threading import Thread
 from collections.abc import Sequence
-from operator import attrgetter
 from pathlib import Path
 from scipy.sparse import issparse
 from annoy import AnnoyIndex
 from tqdm import tqdm
 import numpy as np
 
+from ..data import get_uint_ctype
 
 class NeighbourMatrix(Sequence):
     r"""A matrix A\ :subscript:`ij` where i is the row index of the data point and j
@@ -25,6 +26,12 @@ class NeighbourMatrix(Sequence):
         """The width of the matrix (number of neighbours retrieved)"""
         raise NotImplementedError
 
+    def get_batch(self, idx_seq):
+        """Gets a batch of neighbours corresponding to the provided index sequence.
+
+        Non-optimized version, can be overridden by child classes to be made be efficient"""
+        return [self.__getitem__(item) for item in idx_seq]
+
 class AnnoyKnnMatrix(NeighbourMatrix):
     r"""Neighbouring points are KNN retrieved using an Annoy Index.
 
@@ -45,7 +52,7 @@ class AnnoyKnnMatrix(NeighbourMatrix):
     k = None
 
     def __init__(self, index, nrows, index_path='annoy.index', metric='angular', k=150, search_k=-1,
-                 precompute=False, include_distances=False, verbose=False):
+                 precompute=False, include_distances=False, verbose=False, n_jobs=-1):
         """Constructs an AnnoyKnnMatrix instance from an AnnoyIndex object with given parameters"""
         self.index = index
         self.nrows = nrows
@@ -57,8 +64,15 @@ class AnnoyKnnMatrix(NeighbourMatrix):
         self.include_distances = include_distances
         self.verbose = verbose
         self.precomputed_neighbours = None
+        self.n_jobs = n_jobs
+        self.workers = None
         if precompute:
-            self.precomputed_neighbours = self.get_neighbour_indices()
+            self.precomputed_neighbours = self.get_neighbour_indices(n_jobs=n_jobs)
+        else:
+            if n_jobs and os.cpu_count():
+                # Negative worker counts wrap around to cpu core count, where -1 is one worker/core
+                self.workers = ThreadPoolExecutor(
+                    max_workers=n_jobs if n_jobs > 0 else os.cpu_count() + n_jobs + 1)
 
     def __getstate__(self):
         """ Return object serializable variable dict """
@@ -81,18 +95,18 @@ class AnnoyKnnMatrix(NeighbourMatrix):
 
     @classmethod
     def build(cls, X, path, k=150, metric='angular', search_k=-1, include_distances=False,
-              ntrees=50, build_index_on_disk=True, precompute=False, verbose=1):
+              ntrees=50, build_index_on_disk=True, precompute=False, verbose=1, n_jobs=-1):
         """Builds a new Annoy Index on input data *X*, then constructs and returns an
         AnnoyKnnMatrix object using the newly-built index."""
 
         _validate_knn_shape(X.shape[0], k)
-        index = build_annoy_index(X, path, metric, ntrees, build_index_on_disk, verbose)
+        index = build_annoy_index(X, path, metric, ntrees, build_index_on_disk, verbose, n_jobs)
         return cls(index, X.shape[0], path, metric, k, search_k,
-                   precompute, include_distances, verbose)
+                   precompute, include_distances, verbose, n_jobs)
 
     @classmethod
     def load(cls, index_path, data_shape, k=150, metric='angular', search_k=-1,
-             include_distances=False, precompute=False, verbose=1):
+             include_distances=False, precompute=False, verbose=1, n_jobs=-1):
         """Constructs and returns an AnnoyKnnMatrix object from an existing Annoy Index on disk."""
 
         _validate_knn_shape(data_shape[0], k)
@@ -106,7 +120,7 @@ class AnnoyKnnMatrix(NeighbourMatrix):
         except IOError as err:
             raise IOError("Failed to load annoy index at '%s': file corrupt" % index_path) from err
         return cls(index, data_shape[0], index_path, metric, k, search_k,
-                   precompute, include_distances, verbose)
+                   precompute, include_distances, verbose, n_jobs)
 
     def __len__(self):
         """Number of rows in neighbour matrix"""
@@ -125,11 +139,17 @@ class AnnoyKnnMatrix(NeighbourMatrix):
         return self.index.get_nns_by_item(
             idx, self.k, search_k=self.search_k, include_distances=self.include_distances)
 
-    def get_neighbour_indices(self):
+    def get_neighbour_indices(self, n_jobs=-1):
         """Retrieves neighbours for every row in parallel"""
         if self.precomputed_neighbours is not None:
             return self.precomputed_neighbours
-        return extract_knn(self, verbose=self.verbose)
+        return extract_knn(self, verbose=self.verbose, n_jobs=n_jobs)
+
+    def get_batch(self, idx_seq):
+        """Returns a batch of neighbours based on the index sequence provided."""
+        if self.workers is not None:
+            return list(self.workers.map(self.__getitem__, idx_seq))
+        return super().get_batch(idx_seq)
 
     def save(self, path):
         """Saves internal Annoy index to disk at given path."""
@@ -151,7 +171,8 @@ def _validate_knn_shape(nrows, k):
                             (k={}, rows={}). Lower k to a smaller
                             value.'''.format(k, nrows))
 
-def build_annoy_index(X, path, metric='angular', ntrees=50, build_index_on_disk=True, verbose=1):
+def build_annoy_index(X, path, metric='angular', ntrees=50, build_index_on_disk=True,
+                      verbose=1, n_jobs=-1):
     """ Build a standalone annoy index.
 
     :param array X: numpy array with shape (n_samples, n_features)
@@ -183,21 +204,33 @@ def build_annoy_index(X, path, metric='angular', ntrees=50, build_index_on_disk=
             print('Flattening multidimensional input before building KNN index using Annoy')
         X = X.reshape((X.shape[0], -1))
 
+    batched_retrieval = hasattr(X, 'get_batch')
+
     index = AnnoyIndex(X.shape[1], metric=metric)
     if build_index_on_disk:
         index.on_disk_build(path)
 
-    if issparse(X):
-        for i in tqdm(range(X.shape[0]), disable=verbose < 1):
-            vector = X[i].toarray()[0]
-            index.add_item(i, vector)
+    if not batched_retrieval:
+        if issparse(X):
+            for i in tqdm(range(X.shape[0]), disable=verbose < 1):
+                vector = X[i].toarray()[0]
+                index.add_item(i, vector)
+        else:
+            for i in tqdm(range(X.shape[0]), disable=verbose < 1):
+                vector = X[i]
+                index.add_item(i, vector)
     else:
-        for i in tqdm(range(X.shape[0]), disable=verbose < 1):
-            vector = X[i]
-            index.add_item(i, vector)
+        # Reusing n_jobs for batch_size here - separate config may be better
+        batch_size = n_jobs if n_jobs > 0 else os.cpu_count() + n_jobs + 1
+        for batch_start_idx in tqdm(range(0, X.shape[0], batch_size), disable=verbose < 1):
+            batch_indices = range(batch_start_idx,
+                                  min(batch_start_idx + batch_size, X.shape[0]))
+            batch = X.get_batch(batch_indices)
+            for idx, vector in zip(batch_indices, batch):
+                index.add_item(idx, vector)
 
     try:
-        index.build(ntrees)
+        index.build(ntrees, n_jobs=n_jobs)
     except Exception as e:
         raise Exception("Error building Annoy Index. Passing on_disk_build=False "
                         "may solve the issue, especially on Windows.") from e
@@ -206,7 +239,7 @@ def build_annoy_index(X, path, metric='angular', ntrees=50, build_index_on_disk=
         index.save(path)
     return index
 
-def extract_knn(knn_index, verbose=1):
+def extract_knn(knn_index, verbose=1, n_jobs=-1):
     """Starts multiple threads to retrieve nearest neighbours from a built index in parallel.
 
     :param `NeighbourMatrix` knn_index: Indexable neighbour index. When indexed, returns
@@ -219,7 +252,7 @@ def extract_knn(knn_index, verbose=1):
 
     nrows = len(knn_index)
     try:
-        neighbour_matrix = np.empty((nrows, knn_index.k), dtype=_get_uint_ctype(nrows))
+        neighbour_matrix = np.empty((nrows, knn_index.k), dtype=get_uint_ctype(nrows))
     except (ValueError, MemoryError) as err:
         raise MemoryError("Unable to allocate memory for precomputed KNN matrix. "
                           "Set `precompute` to False or reduce the value for `k`.") from err
@@ -238,11 +271,12 @@ def extract_knn(knn_index, verbose=1):
 
     # Split up the indices and assign a thread for each chunk
     cpus_available = os.cpu_count() or 1
-    chunk_size = max(1, nrows // cpus_available)
+    n_jobs = n_jobs if n_jobs > 0 else cpus_available + n_jobs + 1
+    chunk_size = max(1, nrows // n_jobs)
     data_split_indices = [(min_index, min(nrows, min_index + chunk_size))
                           for min_index in range(0, nrows, chunk_size)]
 
-    progress_counters = [0 for i in range(len(data_split_indices))]
+    progress_counters = [0 for _ in range(len(data_split_indices))]
     thread_pool = [
         Thread(
             target=knn_worker,
@@ -279,13 +313,3 @@ def extract_knn(knn_index, verbose=1):
         for thread in thread_pool:
             thread.join()
         raise
-
-
-def _get_uint_ctype(integer):
-    """Gets smallest possible uint representation of provided integer.
-    Raises ValueError if invalid value provided (negative or above uint64)."""
-    for dtype in np.typecodes["UnsignedInteger"]:
-        min_val, max_val = attrgetter('min', 'max')(np.iinfo(np.dtype(dtype)))
-        if min_val <= integer <= max_val:
-            return dtype
-    raise ValueError("Cannot parse {} as a uint64".format(integer))
